"""Provide hypha-rpc to connecting to Hypha server."""

from .rpc import RPC

from .websocket_client import login, connect_to_server
from .sync_client import login_sync, connect_to_server_sync

__all__ = ["RPC", "login", "connect_to_server", "login_sync", "connect_to_server_sync"]



"""Test the hypha server."""
import pytest
from imjoy_rpc.hypha import login, connect_to_server, login_sync, connect_to_server_sync
from . import WS_SERVER_URL
import numpy as np
import requests
import asyncio


class ImJoyPlugin:
    """Represent a test plugin."""

    def __init__(self, ws):
        """Initialize the plugin."""
        self._ws = ws

    # async def setup(self):
    #     """Set up the plugin."""
    #     await self._ws.log("initialized")

    async def run(self, ctx):
        """Run the plugin."""
        await self._ws.log("hello world")

    async def add(self, data):
        """Add function."""
        return data + 1.0


@pytest.mark.asyncio
async def test_login(socketio_server):
    """Test login to the server."""
    TOKEN = "sf31df234"

    async def callback(context):
        print(f"By passing login: {context['login_url']}")
        loop = asyncio.get_running_loop()
        await loop.run_in_executor(
            None,
            requests.get,
            context["report_url"] + "?key=" + context["key"] + "&token=" + TOKEN,
        )

    # We use ai.imjoy.io to test the login for now
    token = await login(
        {
            "server_url": "https://ai.imjoy.io",
            "login_callback": callback,
            "login_timeout": 3,
        }
    )
    assert token == TOKEN


def test_login_sync(socketio_server):
    """Test login to the server."""
    TOKEN = "sf31df234"

    def callback(context):
        print(f"By passing login: {context['login_url']}")
        requests.get(
            context["report_url"] + "?key=" + context["key"] + "&token=" + TOKEN
        )

    # We use ai.imjoy.io to test the login for now
    token = login_sync(
        {
            "server_url": "https://ai.imjoy.io",
            "login_callback": callback,
            "login_timeout": 3,
        }
    )
    assert token == TOKEN


@pytest.mark.asyncio
async def test_numpy_array_sync(socketio_server):
    """Test numpy array registered in async."""
    ws = connect_to_server_sync(
        {"client_id": "test-plugin", "server_url": WS_SERVER_URL}
    )
    ws.export(ImJoyPlugin(ws))
    workspace = ws.config.workspace
    token = ws.generate_token()

    api = await connect_to_server(
        {
            "client_id": "client",
            "workspace": workspace,
            "token": token,
            "server_url": WS_SERVER_URL,
        }
    )
    plugin = await api.get_service("test-plugin:default")
    result = await plugin.add(2.1)
    assert result == 2.1 + 1.0

    large_array = np.zeros([2048, 2048, 4], dtype="float32")
    result = await plugin.add(large_array)
    np.testing.assert_array_equal(result, large_array + 1.0)


def test_connect_to_server_sync(socketio_server):
    """Test connecting to the server sync."""
    # Now all the functions are sync
    server = connect_to_server_sync(
        {"client_id": "test-plugin", "server_url": WS_SERVER_URL}
    )
    workspace = server.config.workspace
    token = server.generate_token()
    assert workspace and token

    services = server.list_services("public")
    assert isinstance(services, list)

    def hello(name):
        print("Hello " + name)
        return "Hello " + name

    server.register_service(
        {
            "name": "Hello World",
            "id": "hello-world",
            "config": {
                "visibility": "protected",
                "run_in_executor": True,
            },
            "hello": hello,
        }
    )


@pytest.mark.asyncio
async def test_connect_to_server(socketio_server):
    """Test connecting to the server."""
    # test workspace is an exception, so it can pass directly
    ws = await connect_to_server({"name": "my plugin", "server_url": WS_SERVER_URL})
    with pytest.raises(Exception, match=r".*Permission denied for.*"):
        ws = await connect_to_server(
            {"name": "my plugin", "workspace": "test", "server_url": WS_SERVER_URL}
        )
    ws = await connect_to_server({"name": "my plugin", "server_url": WS_SERVER_URL})
    await ws.export(ImJoyPlugin(ws))


async def test_numpy_array(socketio_server):
    """Test numpy array."""
    ws = await connect_to_server(
        {"client_id": "test-plugin", "server_url": WS_SERVER_URL}
    )
    await ws.export(ImJoyPlugin(ws))
    workspace = ws.config.workspace
    token = await ws.generate_token()

    api = await connect_to_server(
        {
            "client_id": "client",
            "workspace": workspace,
            "token": token,
            "server_url": WS_SERVER_URL,
        }
    )
    plugin = await api.get_service("test-plugin:default")
    result = await plugin.add(2.1)
    assert result == 2.1 + 1.0

    large_array = np.zeros([2048, 2048, 4], dtype="float32")
    result = await plugin.add(large_array)
    np.testing.assert_array_equal(result, large_array + 1.0)



"""Provide basic tests for jupyter rpc."""

import datetime
import json
import subprocess
import sys
import time
import uuid

import pytest
import requests
from requests import RequestException
from websocket import create_connection

# The token is written on stdout when you start the notebook
PORT = 9999
BASE_URL = f"http://localhost:{PORT}"


@pytest.fixture(name="jupyter_server")
def jupyter_server_fixture():
    """Start server as test fixture and tear down after test."""
    proc = subprocess.Popen(
        [
            sys.executable,
            "-m",
            "jupyter",
            "notebook",
            "--NotebookApp.token=''",
            "--NotebookApp.disable_check_xsrf=True",
            "--no-browser",
            f"--port={PORT}",
        ]
    )

    timeout = 5
    while timeout > 0:
        try:
            response = requests.get(BASE_URL + "/api/kernels")
            if response.ok:
                print("====> Jupyter server started")
                break
        except RequestException:
            pass
        timeout -= 0.1
        time.sleep(0.1)
    if timeout <= 0:
        raise RuntimeError("Failed to start Jupyter server")
    yield
    proc.terminate()
    proc.wait()


@pytest.fixture(name="websocket_connection")
def websocket_connection_fixture(jupyter_server):
    """Websocket connection fixture."""
    url = BASE_URL + "/api/kernels"
    response = requests.post(url)
    kernel = json.loads(response.text)
    print(f"====> Kernel created {kernel['id']}")
    assert "id" in kernel
    r = requests.get(f'http://127.0.0.1:{PORT}/api/kernels/{ kernel["id"]}')
    assert "id" in r.json()
    print(
        "====> Connecting to ws://127.0.0.1:"
        f'{PORT}/api/kernels/{ kernel["id"]}/channels'
    )
    # Execution request/reply is done on websockets channels
    ws = create_connection(
        f'ws://127.0.0.1:{PORT}/api/kernels/{ kernel["id"]}/'
        "channels?session_id=118d14ea4a234b7b9a8e575f9421de24"
    )
    print(f"====> Websocket connection established {ws.getstatus()}")
    yield ws
    ws.close()


def send_execute_request(code):
    """Send an execution request to the notebook server."""
    msg_type = "execute_request"
    content = {"code": code, "silent": False}
    hdr = {
        "msg_id": uuid.uuid1().hex,
        "username": "test",
        "session": uuid.uuid1().hex,
        "data": datetime.datetime.now().isoformat(),
        "msg_type": msg_type,
        "version": "5.0",
    }
    msg = {"header": hdr, "parent_header": hdr, "metadata": {}, "content": content}
    return msg


def execute(ws, code):
    """Execute python code in a Jupyter notebook."""
    ws.send(json.dumps(send_execute_request(code)))
    # We ignore all the other messages, we just get the code execution output
    msg_type = ""
    while True:
        rsp = json.loads(ws.recv())
        msg_type = rsp["msg_type"]
        if msg_type == "execute_reply":
            if rsp["content"]["status"] == "error":
                print(rsp["content"]["traceback"])
                raise RuntimeError("Failed to execute")
            else:
                assert rsp["content"]["status"] == "ok"
            break
        if msg_type == "stream":
            print(rsp["content"]["text"])


TEST_CODE = """
import asyncio
from imjoy_rpc import connect_to_jupyter, api

fut = connect_to_jupyter()
api.export({})
"""


def test_jupyter_rpc(websocket_connection):
    """Testing imjoy rpc in jupyter notebooks."""
    execute(websocket_connection, TEST_CODE)
